import pytest

from numpy import sqrt

from geo3d import (
    Frame,
    frame_wizard,
    Point,
    Vector,
    UnitFrame,
    RotationMatrix,
    transformation_between_frames,
)


def test_frame_wizard():
    t = frame_wizard([0, 0, 1], [0, 1, 0], "z", "y", [0, 0, 0])
    assert t == UnitFrame


def test_manual_frame_creation():
    rot = RotationMatrix.from_euler_angles("xyz", [90, -45, 45], degrees=True)
    vec = Vector([3, 4, 6])
    f = Frame(rotation_matrix=rot, translation_vector=vec)
    assert f.translation == vec
    assert f.rotation == rot


def test_express_frame_in_frame(example_frames):
    fa,fb,fc = example_frames
    t = fb.express_in_frame(fa)
    assert t.euler_angles("XYZ", degrees=True) == pytest.approx([180, 0, -45])
    assert t.translation.as_array() == pytest.approx([sqrt(2), 0, -4])


def test_transformation_between_frames(example_frames):
    fa,fb,fc = example_frames
    t = transformation_between_frames(fa, fb)
    assert t.euler_angles("XYZ", degrees=True) == pytest.approx([180, 0, -45])
    assert t.translation.as_array() == pytest.approx([1, 1, 4])



from geo3d import Point, Vector, express_point_in_frame, express_points_in_frame
import numpy as np
import pytest


def test_point_from_array():
    p = Point([1, 2, 3])
    assert p[2] == 3


def test_point_from_point():
    p = Point(Point([1, 2, 3]))
    assert p[2] == 3


def test_point_from_vector():
    p = Point(Vector([1, 2, 3]))
    assert p[2] == 3


def test_point_from_tuple():
    p = Point((1, 2, 3))
    assert p[2] == 3


def test_point_from_np():
    p = Point(np.array([1, 2, 3]))
    assert p[2] == 3


def test_point_from_np_bare_nocopy():
    a = np.array([1, 2, 3])
    p = Point.from_array(a, copy=False)
    a[2] = 7
    assert p[2] == 7


def test_point_from_np_bare_copy():
    a = np.array([1, 2, 3])
    p = Point.from_array(a, copy=True)
    a[2] = 7
    assert p[2] == 3


def test_express_point_in_frame(example_frames):
    fa, fb, fc = example_frames
    p0 = (5, 3, 20)
    p1 = (7.07107, 1.41421, -24.00000)
    
    assert Point(p0).express_in_frame(
        fa, original_frame=fb
    ).as_array() == pytest.approx(p1, abs=1e-4)

    assert express_point_in_frame(
        p0, fa, original_frame=fb
    ).as_array() == pytest.approx(p1, abs=1e-4)

    assert Point(p0).express_in_frame(
        fa, original_frame=fb
    ) == Point(p0).express_in_frame(
        fa.express_in_frame(fb)
    )


def test_express_points_in_frame(example_frames):
    fa, fb, fc = example_frames
    p0 = (5, 3, 20)
    p1 = (7.07107, 1.41421, -24.00000)
    
    assert express_points_in_frame(
        np.array([p0,p0]), fa, original_frame=fb
    ) == pytest.approx(np.array([p1,p1]), abs=1e-4)



from geo3d import Vector, Point, express_vector_in_frame
import numpy as np
import pytest

def test_vector_from_array():
    p = Vector([1, 2, 3])
    assert p[2] == 3


def test_vector_from_vector():
    p = Vector(Vector([1, 2, 3]))
    assert p[2] == 3


def test_vector_from_point():
    p = Vector(Point([1, 2, 3]))
    assert p[2] == 3


def test_vector_from_tuple():
    p = Vector((1, 2, 3))
    assert p[2] == 3


def test_vector_from_np():
    p = Vector(np.array([1, 2, 3]))
    assert p[2] == 3


def test_vector_from_np_bare_nocopy():
    a = np.array([1, 2, 3])
    p = Vector.from_array(a, copy=False)
    a[2] = 7
    assert p[2] == 7


def test_vector_from_np_bare_copy():
    a = np.array([1, 2, 3])
    p = Vector.from_array(a, copy=True)
    a[2] = 7
    assert p[2] == 3


def test_express_vector_in_frame(example_frames):
    fa, fb, fc = example_frames
    v0 = (1, 3, 0)
    v1 = (2.82843, -1.41421, 0)

    assert Vector(v0).express_in_frame(
        fa, original_frame=fb
    ).as_array() == pytest.approx(v1, abs=1e-4)

    assert express_vector_in_frame(
        v0, fa, original_frame=fb
    ).as_array() == pytest.approx(v1, abs=1e-4)

    assert Vector(v0).express_in_frame(
        fa, original_frame=fb
    ) == Vector(v0).express_in_frame(
        fa.express_in_frame(fb)
    )



import numpy as np
import math

from .query import centroid, distances_plane_to_points
from .geometry import Plane, Vector, Point, VectorLike
from typing import Sequence, Union, Dict, Tuple


def fit_plane(
    points: Sequence[VectorLike], return_fit_props=False
) -> Union[Plane, Tuple[Plane, Dict]]:
    """Fit plane to points

    Args:
        points: Sequence of points
        return_fit_props: True if dictionary with fit properties should be returned. Defaults to False.

    Returns:
        Plane or tuple (Plane, fit_properties) if return_fit_props==True
    """
    centr = centroid(points).as_array()
    svd = np.linalg.svd(points - centr)
    normal = svd[2][-1]
    plane = Plane(Vector(normal), Point(centr))

    residuals = distances_plane_to_points(plane, points)
    residuals_min = np.min(residuals)
    residuals_max = np.max(residuals)
    flatness = residuals_max - residuals_min

    if return_fit_props:
        fit_props = {
            "rms_error": math.sqrt(np.sum(residuals ** 2) / len(points)),
            "residuals": residuals,
            "residuals_min": residuals_min,
            "residuals_max": residuals_max,
            "flatness": flatness,
        }
        return (plane, fit_props)
    else:
        return plane


import numpy as np

from .geometry import Frame, Point, norm_L2, VectorLike, Plane
from .linalg import dot_vec_vec, norm_L2

from scipy.optimize import minimize
from scipy.spatial.transform import Rotation as R

from typing import Sequence


def distance_between_points(pointA: VectorLike, pointB: VectorLike) -> float:
    """Distance between two points.

    Args:
        pointA: 1st point
        pointB: 2nd point

    Returns:
        Euclidean distance between the two points
    """
    return norm_L2(np.asarray(pointA) - np.asarray(pointB))


def distances_plane_to_points(plane: Plane, points: Sequence[VectorLike]) -> np.ndarray:
    """Shortest distances between plane and points

    The distances are given from the plane to the points, along the plane normal.
    Negative distances indicate points positioned along the negative plane normal.

    Args:
        plane: The queried plane
        points: The sequence of queried points

    Returns:
        np.array of floats: distances for all points
    """

    normal = plane.normal.as_array()
    cartesian_d = dot_vec_vec(normal, plane.point.as_array())
    normal_length = norm_L2(normal)
    return (np.dot(points, normal) - cartesian_d) / normal_length


def centroid(points: Sequence[VectorLike]) -> Point:
    """Centroid for a sequence of points

    Args:
        points: The queried sequence of points

    Returns:
        Point: Centroid point
    """
    return Point(np.sum(points, 0) / len(points))


def minimize_points_to_points_distance(
    groupA: Sequence[VectorLike],
    groupB: Sequence[VectorLike],
    return_report=False,
    method="Powell",
    tol=1e-6,
):
    """Find point group transform to minimize point-group-to-point-group distance.

    Returns a transformation (Frame object) that, if applied to all points in point group
    `groupA`, minimizes the distance between all points in `groupA` an the corresponding
    points in `groupB`.

    Args:
        groupA: Sequence of Points.
        groupB: Sequence of Points (same size as groupA).
        return_report: True if report of minimization algorithm should be returned

    Returns:
        Transformation, or tuple of transformation and minimization report if return_report==True
    """
    # return transform that maps groupA onto groupB with minimum point-to-point distance
    def cost(x):
        [r1, r2, r3, t1, t2, t3] = x
        rot = R.from_rotvec([r1, r2, r3]).as_matrix()
        trans = np.asarray([t1, t2, t3])
        t = Frame(rot, trans)
        c = np.sqrt(
            np.mean(
                np.power(
                    [
                        distance_between_points(pB, Point(pA).transform(t))
                        for (pA, pB) in zip(groupA, groupB)
                    ],
                    2,
                )
            )
        )
        return c

    m = minimize(cost, [0, 0, 0, 0, 0, 0], tol=tol, method=method)
    t = Frame(R.from_rotvec(m["x"][:3]).as_matrix(), m["x"][3:])
    if return_report:
        return t, m
    else:
        return t